<?php

declare(strict_types=1);

namespace FiloBlu\RuleDayFromBirthday\Helper;

use Exception;
use FiloBlu\RuleDayFromBirthday\Api\CustomCouponRepositoryInterface;
use FiloBlu\RuleDayFromBirthday\Api\Data\CustomCouponInterface;
use Magento\Framework\Api\SearchCriteriaBuilder;
use Magento\Framework\App\Helper\AbstractHelper;
use Magento\Framework\App\Helper\Context;
use Magento\Framework\App\ResourceConnection;
use Magento\Framework\EntityManager\MetadataPool;
use Magento\Framework\Exception\InputException;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\Stdlib\DateTime\DateTime;
use Magento\SalesRule\Api\CouponRepositoryInterface;
use Magento\SalesRule\Api\Data\CouponGenerationSpecInterface;
use Magento\SalesRule\Api\Data\CouponInterface;
use Magento\SalesRule\Api\Data\RuleInterface;
use Magento\SalesRule\Api\RuleRepositoryInterface;
use Magento\SalesRule\Model\Coupon as CouponModel;
use Magento\SalesRule\Model\Data\CouponGenerationSpec;
use Magento\SalesRule\Model\Service\CouponManagementService;
use Magento\Store\Model\StoreManagerInterface;
use Zend_Db;

/**
 * @class Data
 * @package FiloBlu\BirthdayCoupon\Helper
 */
class Data extends AbstractHelper
{

    /** @var string */
    const XML_BIRTHDAY_COUPON_DAYS_OFFSET = 'filoblu_ruledayfrombirthday_section/birthday_coupon_id/birthday_advance_offset';

    /** @var array */
    protected $birthdayRule = [];
    /** @var int */
    protected $maxValidityDays, $couponLength;
    /** @var string */
    protected $couponPrefix, $couponSuffix;

    /**
     * @var ResourceConnection
     */
    protected $resourceConnection;
    /**
     * @var RuleRepositoryInterface
     */
    protected $ruleRepository;
    /**
     * @var StoreManagerInterface
     */
    protected $storeManager;
    /**
     * @var CouponManagementService
     */
    private $couponManagement;
    /**
     * @var CouponGenerationSpec
     */
    private $couponGenerationSpec;
    /**
     * @var CouponModel
     */
    private $couponModel;
    /**
     * @var DateTime
     */
    private $dateTime;
    /**
     * @var CustomCouponRepositoryInterface
     */
    private $customCouponRepository;
    /**
     * @var MetadataPool
     */
    private $metadataPool;
    /**
     * @var CouponRepositoryInterface
     */
    private $couponRepository;
    /**
     * @var SearchCriteriaBuilder
     */
    private $searchCriteriaBuilder;

    /**
     * @param Context $context
     * @param DateTime $dateTime
     * @param ResourceConnection $resourceConnection
     * @param StoreManagerInterface $storeManager
     * @param RuleRepositoryInterface $ruleRepository
     * @param CouponManagementService $couponManagement
     * @param CouponGenerationSpec $couponGenerationSpec
     * @param CouponModel $couponModel
     * @param CustomCouponRepositoryInterface $customCouponRepository
     * @param CouponRepositoryInterface $couponRepository
     * @param MetadataPool $metadataPool
     * @param SearchCriteriaBuilder $searchCriteriaBuilder
     */
    public function __construct(
        Context                         $context,
        DateTime                        $dateTime,
        ResourceConnection              $resourceConnection,
        StoreManagerInterface           $storeManager,
        RuleRepositoryInterface         $ruleRepository,
        CouponManagementService         $couponManagement,
        CouponGenerationSpecInterface   $couponGenerationSpec,
        CouponModel                     $couponModel,
        CustomCouponRepositoryInterface $customCouponRepository,
        CouponRepositoryInterface       $couponRepository,
        MetadataPool                    $metadataPool,
        SearchCriteriaBuilder           $searchCriteriaBuilder
    ) {
        $this->resourceConnection = $resourceConnection;
        $this->dateTime = $dateTime;
        $this->ruleRepository = $ruleRepository;
        $this->storeManager = $storeManager;
        $this->couponManagement = $couponManagement;
        $this->couponGenerationSpec = $couponGenerationSpec;
        $this->couponModel = $couponModel;
        $this->customCouponRepository = $customCouponRepository;
        $this->couponRepository = $couponRepository;
        $this->metadataPool = $metadataPool;
        $this->searchCriteriaBuilder = $searchCriteriaBuilder;
        parent::__construct($context);
    }

    /**
     * @return int
     */
    public function getCouponCreationDaysOffset(): int
    {
        return (int)$this->scopeConfig->getValue(self::XML_BIRTHDAY_COUPON_DAYS_OFFSET) ?? 0;
    }

    /**
     * @param int $storeId
     * @return RuleInterface|null
     * @throws LocalizedException
     * @throws Exception
     */
    public function getBirthdayCouponRule(int $storeId = 0): ?RuleInterface
    {
        if (isset($this->birthdayRule[$storeId])) {
            return $this->birthdayRule[$storeId];
        }

        $connection = $this->resourceConnection->getConnection();
        $salesRuleTable = $this->resourceConnection->getTableName('salesrule');
        $salesRuleWebsiteTable = $this->resourceConnection->getTableName('salesrule_website');
        $salesRuleCouponTable = $this->resourceConnection->getTableName('filoblu_sales_rule_coupon');
        $metadata = $this->metadataPool->getMetadata(RuleInterface::class);
        $linkField = $metadata->getLinkField();
        $websiteId = $this->storeManager->getStore($storeId)->getWebsiteId();

        $sql = "
            SELECT *
            FROM $salesRuleTable as `sr`
            LEFT JOIN $salesRuleWebsiteTable as `srw`
                ON `sr`.`$linkField` = `srw`.`$linkField`
            LEFT JOIN $salesRuleCouponTable as `src`
                ON `sr`.`rule_id` = `src`.`rule_id`
            WHERE
                `sr`.`is_active` = ?
            AND
                `srw`.`website_id` = ?
            AND
                `src`.`type` = ?
            ORDER BY `sr`.`sort_order` ASC
        ";

        $ruleId = $connection->fetchOne($sql, [1, $websiteId, 'birthday_coupon']);
        if (empty($ruleId) || !is_numeric($ruleId)) {
            return null;
        }

        return $this->birthdayRule[$storeId] = $this->ruleRepository->getById($ruleId);
    }

    /**
     * @param string $date
     * @return array
     */
    public function getSubscribersForDate(string $date): array
    {
        $formattedDate = $formattedDateNoZero = $this->getMonthDayFromDate($date);
        if ($formattedDate[1] === '0') {
            $formattedDateNoZero = str_replace('-0', '-', $formattedDateNoZero);
        }
        $connection = $this->resourceConnection->getConnection();
        $subscriberInfoTable = $this->resourceConnection->getTableName('filoblu_newsletter_subscriber_info');
        $subscriberTable = $this->resourceConnection->getTableName('newsletter_subscriber');
        $sql = "
            SELECT *
            FROM {$subscriberInfoTable} as `fnsi`
            LEFT JOIN {$subscriberTable} as `ns`
                ON `fnsi`.`subscriber_id` = `ns`.`subscriber_id`
            WHERE `fnsi`.`dob` like ?
               OR `fnsi`.`dob` like ?

        ";

        return $connection->fetchAll($sql, ["%$formattedDate%", "%$formattedDateNoZero%"], Zend_Db::FETCH_ASSOC);
    }

    /**
     * @param string $date
     * @return string
     */
    public function getMonthDayFromDate(string $date): string
    {
        $time = strtotime($date);
        $month = date("m", $time);
        $day = date("d", $time);

        return '-' . $month . '-' . $day;
    }

    /**
     * @param RuleInterface $rule
     * @param $quantity
     * @return CouponInterface[]
     * @throws InputException
     * @throws LocalizedException
     */
    public function generateCoupon(RuleInterface $rule, $quantity): array
    {
        /** @var CustomCouponInterface $customCoupon */
        $customCoupon = $this->customCouponRepository->getByKey($rule->getRuleId(), 'rule_id');

        $this->couponGenerationSpec
            ->setRuleId($rule->getRuleId())
            ->setLength($customCoupon->getLength())
            ->setFormat($customCoupon->getFormat())
            ->setPrefix($customCoupon->getPrefix())
            ->setSuffix($customCoupon->getSuffix())
            ->setDelimiterAtEvery($customCoupon->getDash())
            ->setQuantity($quantity);

        $daysValidity = $customCoupon->getExpirationDays();
        $today = $this->dateTime->date();
        $couponCodes = $this->couponManagement->generate($this->couponGenerationSpec);
        foreach ($couponCodes as $couponCode) {
            /** @var CouponInterface $coupon */
            $searchCriteria = $this->searchCriteriaBuilder
                ->addFilter('code', $couponCode)
                ->create();
            $couponSearchResult = $this->couponRepository->getList($searchCriteria);
            $coupon = current($couponSearchResult->getItems());
            $coupon->setCreatedAt($today);
            if ($daysValidity) {
                $expDate = date('Y-m-d H:i:s', strtotime("+ $daysValidity days"));
                $coupon->setExpirationDate($expDate);
            }

            // it only works if I add use_auto_generation to the salesrule,
            // also, the coupon itself would be applicable only if the use_auto_generation in flagged
            // I'd go with Specific Coupon Type
            $this->couponRepository->save($coupon);
        }

        return $couponCodes;
    }

    /**
     * @param $rule_id
     * @return int
     * @throws Exception
     */
    public function getSalesRuleLinkIdByRuleId($rule_id): int
    {
        $connection = $this->resourceConnection->getConnection();
        $linkField = $this->metadataPool->getMetadata(RuleInterface::class)->getLinkField();
        $salesRuleTable = $this->resourceConnection->getTableName('salesrule');
        $sql = "
            SELECT $linkField
            FROM {$salesRuleTable} as `sr`
            WHERE `sr`.`rule_id` like ?
        ";

        return (int)$connection->fetchOne($sql, [$rule_id]);
    }
}
